'' =================================================================================================
''
''   File....... jm_i2c.spin
''   Purpose.... Low-level I2C routines (requires pull-ups on SCL and SDA)
''   Author..... Jon "JonnyMac" McPhalen
''               Copyright (c) 2009-2013 Jon McPhalen
''               -- elements inspired by code from Mike Green
''   E-mail.....  
''   Started.... 28 JUL 2009
''   Updated.... 07 APR 2013
''
'' =================================================================================================


con { fixed io pins }

  RX1    = 31                                                   ' programming / debug port
  TX1    = 30
  
  EE_SDA = 29                                                   ' eeprom
  EE_SCL = 28

 
con

  #0, ACK, NAK


var

  long  scl                                                     ' buss pins
  long  sda
    

pub setup

'' Setup I2C using Propeller EEPROM pins

  setupx(EE_SCL, EE_SDA)
         

pub setupx(sclpin, sdapin)

'' Define I2C SCL (clock) and SDA (data) pins

  longmove(@scl, @sclpin, 2)                                    '  copy pins
  dira[scl] := 0                                                '  float to pull-up
  outa[scl] := 0                                                '  write 0 to output reg
  dira[sda] := 0
  outa[sda] := 0

  repeat 9                                                      ' reset device
    dira[scl] := 1
    dira[scl] := 0
    if (ina[sda])
      quit
  
    
pub wait(id) | ackbit

'' Waits for I2C device to be ready for new command

  repeat
    start
    ackbit := write(id)
  until (ackbit == ACK)


pub start

'' Create I2C start sequence
'' -- will wait if I2C buss SDA pin is held low

  dira[sda] := 0                                                ' float SDA (1)
  dira[scl] := 0                                                ' float SCL (1)
  repeat while (ina[scl] == 0)                                  ' allow "clock stretching"

  dira[sda] := 1                                                ' SDA low (0)
  dira[scl] := 1                                                ' SCL low (0)

  
pub write(i2cbyte) | ackbit

'' Write byte to I2C buss
'' -- leaves SCL low

  i2cbyte := (i2cbyte ^ $FF) << 24                              ' move msb (bit7) to bit31
  repeat 8                                                      ' output eight bits
    dira[sda] := i2cbyte <-= 1                                  ' send msb first
    dira[scl] := 0                                              ' SCL high (float to p/u)
    dira[scl] := 1                                              ' SCL low

  dira[sda] := 0                                                ' relase SDA to read ack bit
  dira[scl] := 0                                                ' SCL high (float to p/u)  
  ackbit := ina[sda]                                            ' read ack bit
  dira[scl] := 1                                                ' SCL low

  return (ackbit & 1)


pub read(ackbit) | i2cbyte

'' Read byte from I2C buss

  dira[sda] := 0                                                ' make sda input

  repeat 8
    dira[scl] := 0                                              ' SCL high (float to p/u)
    i2cbyte := (i2cbyte << 1) | ina[sda]                        ' read the bit
    dira[scl] := 1                                              ' SCL low
                             
  dira[sda] := !ackbit                                          ' output ack bit 
  dira[scl] := 0                                                ' clock it
  dira[scl] := 1

  return (i2cbyte & $FF)


pub stop

'' Create I2C stop sequence 

  dira[sda] := 1                                                ' SDA low
  dira[scl] := 0                                                ' float SCL
  repeat
  until (ina[scl] == 1)                                         ' hold for clock stretch
  
  dira[sda] := 0                                                ' float SDA


 ' // -------------------- Extras --------------------

 
PUB ReStart                ' SDA goes HIGH to LOW with SCL HIGH
   outa[SDA]~~                         ' Initially drive SDA HIGH
   dira[SDA]~~
   outa[SCL]~~                         ' Initially drive SCL HIGH
   'waitcnt(clkfreq / I2CDelay + cnt)
   outa[SDA]~                          ' Now drive SDA LOW
   'waitcnt(clkfreq / I2CDelay + cnt)
   outa[SCL]~                          ' Leave SCL LOW

PUB devicePresent(deviceAddress) : ackbit
  ' send the deviceAddress and listen for the ACK
   Start
   ackbit := Write(deviceAddress | 0)
   Stop
   if ackbit == ACK
     return true
   else
     return false

PUB readLocation(device_address, register) : value
    start                    ' 1. Send a start sequence
    write(device_address | 0) ' 2. Send 0xC0 ( I2C address of the CMPS03 with the R/W bit low (even address)
    write(register)           ' 3. Send 0x01 (Internal address of the bearing register)  
    start                    ' 4. Send a start sequence again (repeated start) 
    write(device_address | 1) ' 5. Send 0xC1 ( I2C address of the CMPS03 with the R/W bit high (odd address) 
    value := read(NAK)        ' 6. Read data byte from CMPS03 
    stop                     ' 7. Send the stop sequence. 
    return value                  ' ^- Copied From (http://www.robot-electronics.co.uk/acatalog/I2C_Tutorial.html)
     
PUB writeLocation(device_address, register, value)
    start                    ' 1. Send a start sequence         
    write(device_address)     ' 2. Send the I2C address of the slave with the R/W bit low (even address) 
    write(register)           ' 3. Send the internal register number you want to write to     
    write(value)              ' 4. Send the data byte      5. [Optionally, send any further data bytes]                   
    stop                     ' 6. Send the stop sequence.
        
PUB readLocation16(device_address, register) : value
    start
    write(device_address | 0)
    write(register)
    restart
    write(device_address | 1)  

    'value := read(SCL,ACK)
    'value <<= 8

    value := read(ACK)
    value |= read(ACK) << 8

    'readkeyvar := i2c2.i2cRead(i2cSCL2, 0)        'read first bit field 0 - 7
    'readkeyvar |= i2c2.i2cRead(i2cSCL2, 0) << 8   'read second bit field 8 - 15
        
    value |= (read(NAK) & $ff)
    stop
    return value
     
PUB readLocation24(device_address, register) : value
    start
    write(device_address | 0)
    write(register)
    restart
    write(device_address | 1)  
    value := read(ACK)
    value <<= 8
    value |= (read(ACK) & $ff)
    value <<= 8
    value |= (read(NAK) & $ff)
    stop
    return value
     
  

dat

{{

  Terms of Use: MIT License

  Permission is hereby granted, free of charge, to any person obtaining a copy of this
  software and associated documentation files (the "Software"), to deal in the Software
  without restriction, including without limitation the rights to use, copy, modify,
  merge, publish, distribute, sublicense, and/or sell copies of the Software, and to
  permit persons to whom the Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be included in all copies
  or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
  INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
  PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
  CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE
  OR THE USE OR OTHER DEALINGS IN THE SOFTWARE. 

}}